
var BingMapView = new function() {

    //
    // Our map. Single instance for the page and corresponding Java BingMapView object
    //
    var map;

    //
    // Array of map event handlers
    //
    var mapEventHandlers;

    //
    // Entity map.
    // Maps entity ids (generated in Java) to  a JavaScript entity (PushPin, Polyline, Polygon, TileLayer)
    //
    var entityMap;

    //
    // Whether load of the Bing Maps AJAX SDK failed
    var scriptLoadFailed = false;

    //
    // Utility functions
    //

    //
    // Converts the location format used in serialization to Microsoft.Maps.Location object
    //
    function convertLocation(parsedLocation) {
        // Location is serialized as a simple { latitude; longitude } struct/class
        // Convert that to a Microsoft.Maps.Location object
        return new Microsoft.Maps.Location(parsedLocation.latitude, parsedLocation.longitude);
    }

    //
    // Parses MapOptions and ViewOptions JSON strings received from Java
    //
    function parseMapAndViewOptions(serializedOptions) {
        var optionsObject = JSON.parse(serializedOptions);

        if (typeof optionsObject.bounds !== "undefined") {
            optionsObject.bounds =
                new Microsoft.Maps.LocationRect.fromEdges(optionsObject.bounds.north,
                                                          optionsObject.bounds.west,
                                                          optionsObject.bounds.south,
                                                          optionsObject.bounds.east,
                                                          0,
                                                          Microsoft.Maps.AltitudeReference.ground);
        }

        if (typeof optionsObject.center !== "undefined") {
            optionsObject.center = convertLocation(optionsObject.center);
        }

        if (typeof optionsObject.centerOffset !== "undefined") {
            optionsObject.centerOffset = new Microsoft.Maps.Point(optionsObject.centerOffset.x,
                                                                  optionsObject.centerOffset.y);
        }

        if (typeof optionsObject.labelOverlay !== "undefined") {
            console.log("setting labelOverlay to ordinal:" + optionsObject.labelOverlay);
            switch (optionsObject.labelOverlay) {
                case 0:
                    console.log("setting labelOverlay to hidden");
                    optionsObject.labelOverlay = Microsoft.Maps.LabelOverlay.hidden
                    break;

                case 1:
                    console.log("setting labelOverlay to visible");
                    optionsObject.labelOverlay = Microsoft.Maps.LabelOverlay.visible;
                    break;

                default:
                    console.log("setting labelOverlay to visible1");
                    optionsObject.labelOverlay = Microsoft.Maps.LabelOverlay.visible;
            }
        }

        if (typeof optionsObject.mapTypeId !== "undefined") {
            console.log("setting maptypeid to ordinal:" + optionsObject.mapTypeId);
            switch (optionsObject.mapTypeId) {
                case 0:
                    optionsObject.mapTypeId = Microsoft.Maps.MapTypeId.road;
                    break;

                case 1:
                    optionsObject.mapTypeId = Microsoft.Maps.MapTypeId.auto;
                    break;

                case 2:
                    optionsObject.mapTypeId = Microsoft.Maps.MapTypeId.aerial;
                    break;

                case 3:
                    optionsObject.mapTypeId = Microsoft.Maps.MapTypeId.birdseye;
                    break;

                default:
                    optionsObject.mapTypeId = Microsoft.Maps.MapTypeId.auto;
            }
        }

        return optionsObject;
    }

    //
    //  Adjust default options values on map creation
    //
    function adjustMapAndViewOptionsDefault(options) {
        //
        // Options NOT exposed to apps where our defaults are different than the AJAX control defaults
        //
        options.theme = new Microsoft.Maps.Themes.BingTheme();
        options.enableClickableLogo = false;
        options.enableSearchLogo = false;
        options.showCopyright = false;
        options.showDashboard = false;
        options.showMapTypeSelector = false;
        options.showScalebar = false;

        //
        // Options exposed to apps where our defaults are different than the AJAX control defaults
        //   Set our defaults if not specified by the app
        //
        if (typeof options.useInertia === 'undefined') {
            options.useInertia = false;
        }

        if (((typeof options.useInertia === 'undefined') || (options.useInertia === true)) &&
             (typeof options.inertiaIntensity === 'undefined')) {
            options.inertiaIntensity = 0.8;
        }

        if (typeof options.tileBuffer === 'undefined') {
            options.tileBuffer = 2;
        }
        return options;
    }

    //
    // Parses PushPinOptions
    //
    function parsePinOptions(serializedOptions) {
        var optionsObject = JSON.parse(serializedOptions);

        if (typeof optionsObject.anchor !== "undefined") {
            optionsObject.anchor = new Microsoft.Maps.Point(optionsObject.anchor.x,
                                                            optionsObject.anchor.y);
        }

        if (typeof optionsObject.textOffset !== "undefined") {
            optionsObject.textOffset = new Microsoft.Maps.Point(optionsObject.textOffset.x,
                                                                optionsObject.textOffset.y);
        }
        return optionsObject;
    }

    //
    // Parses PolylineOptions
    //
    function parsePolylineOptions (serializedOptions) {
        var optionsObject = JSON.parse(serializedOptions);

        if (typeof optionsObject.strokeColor !== "undefined") {
            optionsObject.strokeColor = new Microsoft.Maps.Color(optionsObject.strokeColor.a,
                                                                 optionsObject.strokeColor.r,
                                                                 optionsObject.strokeColor.g,
                                                                 optionsObject.strokeColor.b);
        }
        return optionsObject;
    }

    //
    // Parses PolylgonOptions
    //
    function parsePolygonOptions (serializedOptions) {
        var optionsObject = JSON.parse(serializedOptions);

        if (typeof optionsObject.strokeColor !== "undefined") {
            optionsObject.strokeColor = new Microsoft.Maps.Color(optionsObject.strokeColor.a,
                                                                 optionsObject.strokeColor.r,
                                                                 optionsObject.strokeColor.g,
                                                                 optionsObject.strokeColor.b);
        }

        if (typeof optionsObject.fillColor !== "undefined") {
            optionsObject.fillColor = new Microsoft.Maps.Color(optionsObject.fillColor.a,
                                                               optionsObject.fillColor.r,
                                                               optionsObject.fillColor.g,
                                                               optionsObject.fillColor.b);
        }
        return optionsObject;
    }

    //
    // Parses a location array
    //
    function parseLocationArray(serializedLocationArray) {
        var parsedArray = JSON.parse(serializedLocationArray);
        var locationObjectArray = new Array();

        for (var i = 0; i < parsedArray.length; i++)
        {
            locationObjectArray.push(convertLocation(parsedArray[i]));
        }
        return locationObjectArray;
    }

    //
    // Parses tile layer options
    //
    function parseTileLayerOptions(serializedOptions) {
        var optionsObject = JSON.parse(serializedOptions);

        if (typeof optionsObject.animationDisplay !== "undefined") {
            switch (optionsObject.animationDisplay) {
                case 0:
                    console.log("animationDisplay setting to auto");
                    optionsObject.animationDisplay = Microsoft.Maps.AnimationVisibility.auto;
                    break;

                case 1:
                    console.log("animationDisplay setting to hide");
                    optionsObject.animationDisplay = Microsoft.Maps.AnimationVisibility.hide;
                    break;

                case 2:
                    console.log("animationDisplay setting to show");
                    optionsObject.animationDisplay = Microsoft.Maps.AnimationVisibility.show;
                    break;

                default:
                    console.log("animationDisplay setting to auto1");
                    optionsObject.animationDisplay = Microsoft.Maps.AnimationVisibility.auto;
            }
        }
        return optionsObject;
    }

    //
    // Map functions
    //

    //
    // Creates a map, given a set of options (MapOptions and ViewOptions combined)
    //
    this.createMap = function (serializedOptions) {
        console.log("Create map called");
        if (typeof map !== 'undefined' && map !== null) {
                this.deleteMap();
        }
        var allOptions = parseMapAndViewOptions(serializedOptions);
        allOptions = adjustMapAndViewOptionsDefault(allOptions);

        map = new Microsoft.Maps.Map(document.getElementById("mapDiv"), allOptions);

        // Make the map object background transparent (SDK will ignore alpha of an specified Microsoft.Maps.Color object)
        // Actual background will be handled in the containing Android WebView
        var elements = document.getElementsByClassName("MicrosoftMap");
        if (typeof (elements) != 'undefined' && elements != null && elements.length > 0) {
            var outerDiv = elements[0];
            outerDiv.style.cssText = outerDiv.style.cssText.replace(/background-color:[^;]*;/gi,
                                                                    "background-color: transparent !important;");
        }

        entityMap = {};
        map.entities.clear();

        // Add handlers to the map events we are interested in
        // These will be fired back to Java
        mapEventHandlers = [];
        mapEventHandlers.push(Microsoft.Maps.Events.addThrottledHandler(map, "viewchange", onViewChange, 100));
        mapEventHandlers.push(Microsoft.Maps.Events.addHandler(map, "viewchangestart", onViewChangeStart));
        mapEventHandlers.push(Microsoft.Maps.Events.addHandler(map, "viewchangeend", onViewChangeEnd));
        mapEventHandlers.push(Microsoft.Maps.Events.addHandler(map, "maptypechanged", onMapTypeChanged));
        mapEventHandlers.push(Microsoft.Maps.Events.addHandler(map, "click", onMapClicked));
        mapEventHandlers.push(Microsoft.Maps.Events.addHandler(map, "imagerychanged", onImageryChanged));
        mapEventHandlers.push(Microsoft.Maps.Events.addHandler(map, "targetviewchanged", onTargetViewChanged));
        mapEventHandlers.push(Microsoft.Maps.Events.addHandler(map, "tiledownloadcomplete", onTileDownloadComplete));
    }

    //
    // Deletes a map
    //
    this.deleteMap = function () {
        if (typeof map !== 'undefined' && map !== null) {
            if (mapEventHandlers) {
                for (handler in mapEventHandlers) {
                    Microsoft.Maps.Events.removeHandler(handler);
                }
            }
            mapEventHandlers = [];
            map.dispose();
            map = null;
            entityMap = {};
        } else {
            console.warn("Trying to delete map null or not defined");
        }
    }

    //
    // Map event handlers
    //
    function onViewChange(e) {
        if (window.BingMapViewJavaScriptBridge) {
            window.BingMapViewJavaScriptBridge.onViewChange();
        }
    }

    function onViewChangeStart(e) {
        if (window.BingMapViewJavaScriptBridge) {
            window.BingMapViewJavaScriptBridge.onViewChangeStart();
        }
    }

    function onViewChangeEnd(e) {
        if (window.BingMapViewJavaScriptBridge) {
            // At the end of a view change, send some basic current map info to the Java view
            var c = map.getCenter();
            var z = map.getZoom();
            var b = map.getBounds();
            window.BingMapViewJavaScriptBridge.onViewChangeEnd(c.latitude, c.longitude, z,
                                                               b.getNorth(), b.getEast(), b.getSouth(), b.getWest(),
                                                               map.getImageryId());
        }
    }

    function onMapTypeChanged(e) {
        if (window.BingMapViewJavaScriptBridge) {
            window.BingMapViewJavaScriptBridge.onMapTypeChanged(map.getImageryId());
        }
    }

    function onMapClicked(e) {
        if (window.BingMapViewJavaScriptBridge) {
            if (e.targetType == "map") {
                window.BingMapViewJavaScriptBridge.onMapClicked(e.pageX, e.pageY);
            }
        }
    }

    function onViewChange(e) {
        console.log('onViewChange');
        if (window.BingMapViewJavaScriptBridge) {
            window.BingMapViewJavaScriptBridge.onViewChange();
        }
    }

    function onImageryChanged(e) {
        console.log('onImageryChanged');
    }

    function onTargetViewChanged(e) {
        console.log('onTargetViewChanged');
    }

    function onTileDownloadComplete(e) {
        console.log('onTileDownloadComplete');
    }

    //
    // Set options on an already created map
    //   NOTE - not all options are settable after creation
    //         see Microsoft.Maps.MapOptions for details
    //
    this.setOptions = function (serializedMapOptions) {
        if (typeof map !== "undefined") {
            map.setOptions(parseMapAndViewOptions(serializedMapOptions));
        }
    }

    //
    // Sets the map view based on the given set of options
    //
    this.setView = function (serializedViewOptions) {
        if (typeof map !== "undefined") {
            map.setView(parseMapAndViewOptions(serializedViewOptions));
        }
    }

    //
    // Restricts the map zoom to a range
    //
    this.restrictZoom = function (minLevel, maxLevel) {
        if (typeof map !== "undefined") {
            console.log("Restricting map zoom level. min:" + minLevel + " max:" + maxLevel);
            map.restrictZoom(minLevel, maxLevel);
        }
    }

    //
    // Entity functions
    //

    //
    // Add a newly created entity to the proper maps, as well as adding a click handler
    //
    function addEntity(entity, id) {
        map.entities.push(entity);
        entity.clickHandler = Microsoft.Maps.Events.addHandler(entity, "click", function(e) {
            if (window.BingMapViewJavaScriptBridge) {
                window.BingMapViewJavaScriptBridge.onEntityClicked(id);
            }
        });
        entityMap[id] = entity;
    }

    //
    // Destroy entity
    //
    this.destroyEntity = function (entityId) {
        var entity = entityMap[entityId];
        if (typeof entity === "undefined") {
            console.error("Trying to destroy entity not mapped. Id:" + entityId);
        } else {
            map.entities.remove(entity);
            if (entity.clickHandler) {
                Microsoft.Maps.Events.removeHandler(entity.clickHandler);
            }
            delete entityMap[entityId];
            console.log('Entity map length: ' + Object.keys(entityMap).length);
        }
    }

    //
    // Pin functions
    //

    //
    // Create pin
    //
    this.createPin = function (pinId, serializedLocation, serializedPinOptions) {
        var location =  convertLocation(JSON.parse(serializedLocation));
        var options = parsePinOptions(serializedPinOptions);
        var pushpin= new Microsoft.Maps.Pushpin(location, options);
        addEntity(pushpin, pinId);
    }

    //
    // Set pin location
    //
    this.setPinLocation = function (pinId, serializedLocation) {
        var pin = entityMap[pinId];
        if (typeof pin === 'undefined') {
            console.error("Trying to set location on pin not mapped. Id:" + pinId);
        } else {
            var location =  convertLocation(JSON.parse(serializedLocation));
            pin.setLocation(location);
            //console.log("Set pin location. Id:" + pinId + " Latitude:" + location.latitude + " Longitude:" + location.longitude);
        }
    }

    //
    // Set pin options
    //
    this.setPinOptions = function (pinId, serializedPinOptions) {
        var pin = entityMap[pinId];
        if (typeof pin === 'undefined') {
            console.error("Trying to set options on pin not mapped. Id:" + pinId);
        } else {
            var options = parsePinOptions(serializedPinOptions);
            pin.setOptions(options);
            //console.log("Set pin options. Id:" + pinId);
        }
    }

    //
    // Polyline functions
    //

    //
    // Create Polyline
    //
    this.createPolyline = function (polylineId, serializedLocations, serializedPolylineOptions) {
        var locations = parseLocationArray(serializedLocations);
        var options = parsePolylineOptions(serializedPolylineOptions);
        var polyline = new Microsoft.Maps.Polyline(locations, options);
        addEntity(polyline, polylineId);
    }

    //
    // Set Polyline locations
    //
    this.setPolylineLocations = function (polylineId, serializedLocations) {
        var polyline = entityMap[polylineId];
        if (typeof polyline === 'undefined') {
            console.error("Trying to set locations on polyline not mapped. Id:" + polylineId);
        } else {
            var locations =  parseLocationArray(serializedLocations);
            polyline.setLocations(locations);
            console.log("Set polyline locations. Id:" + polylineId + " Length:" + locations.length);
        }
    }

    //
    // Set Polyline options
    //
    this.setPolylineOptions =  function (polylineId, serializedPolylineOptions) {
        var polyline = entityMap[polylineId];
        if (typeof polyline === 'undefined') {
            console.error("Trying to set options on polyline not mapped. Id:" + polylineId);
        } else {
            var options = parsePolylineOptions(serializedPolylineOptions);
            polyline.setOptions(options);
            console.log("Set polyline options. Id:" + polylineId);
        }
    }

    //
    // Polygon functions
    //

    //
    // Create Polygon
    //
    this.createPolygon = function (polygonId, serializedLocations, serializedPolygonOptions) {
        var locations =  parseLocationArray(serializedLocations);
        var options = parsePolygonOptions(serializedPolygonOptions);
        var polygon= new Microsoft.Maps.Polygon(locations, options);
        addEntity(polygon, polygonId);
    }

    //
    // Set Polygon locations
    //
    this.setPolygonLocations = function (polygonId, serializedLocations) {
        var polygon = entityMap[polygonId];
        if (typeof polygon === 'undefined') {
            console.error("Trying to set locations on polygon not mapped. Id:" + polygonId);
        } else {
        var locations =  parseLocationArray(serializedLocations);
            polygon.setLocations(locations);
            console.log("Set polygon locations. Id:" + polygonId + " Length:" + locations.length);
        }
    }

    //
    // Set Polygon options
    //
    this.setPolygonOptions = function (polygonId, serializedPolylineOptions) {
        var polygon = entityMap[polygonId];
        if (typeof polygon === 'undefined') {
            console.error("Trying to set options on polygon not mapped. Id:" + polygonId);
        } else {
        var options = parsePolygonOptions(serializedPolylineOptions);
            polygon.setOptions(options);
            console.log("Set polygon options. Id:" + polylineId);
        }
    }

    //
    // Tile Layer functions
    //
    this.createTileLayer = function (id, serializedOptions) {
        var options = parseTileLayerOptions(serializedOptions);
        //
        // options.mercator comes with a serialized TileSourceOptions, not TileSource
        // options.mercator.uriConstructor has an app provided tile path Uri template
        //
        // In order to achieve what we want we need to replace the app provided tile uri template with a function pointer
        // The provided function will then produce the individual tiles Uris, using the app provided template and tile info (x, y, zoom level)
        //  We store the app provided Uri template as an additional property - clientUriTemplate - in the created TileLayer
        //
        var clientUriTemplate = options.mercator.uriConstructor;
        var tileLayer;
        options.mercator.uriConstructor = function (tile) {
            var formatted = tileLayer.clientUriTemplate.replace(/{x}/gi, tile.x)
                                                       .replace(/{y}/gi, tile.y)
                                                       .replace(/{z}/gi, tile.levelOfDetail);
            return formatted;
        }
        options.mercator = new Microsoft.Maps.TileSource(options.mercator);
        tileLayer = new Microsoft.Maps.TileLayer(options);
        tileLayer.clientUriTemplate = clientUriTemplate;
        addEntity(tileLayer, id);
    }

    //
    // Set Tile Layer options
    //
    this.setTileLayerOptions = function (id, serializedTileLayerOptions) {
        var tileLayer = entityMap[id];
        if (typeof tileLayer === 'undefined') {
            console.error("Trying to set options on tile layer not mapped. Id:" + id);
        } else {
            var options = parseTileLayerOptions(serializedTileLayerOptions);
            if (options.mercator) {
                if (options.mercator.uriConstructor) {
                    // If we are given a new app defined Uri template, just store it in our tile layer and replace it with our function
                    tileLayer.clientUriTemplate = options.mercator.uriConstructor;
                    options.mercator.uriConstructor = tileLayer.getTileSource().getUriConstructor();
                    options.mercator = new Microsoft.Maps.TileSource(options.mercator);
                }
            }
            tileLayer.setOptions(options);
            console.log("Set tile layer options. Id:" + id);
        }
    }

    //
    // Load BingTheme
    //
    this.loadBingTheme = function(){
        if (!scriptLoadFailed) {
            console.log("Starting BingTheme load");
            Microsoft.Maps.loadModule('Microsoft.Maps.Themes.BingTheme', { callback: themesModuleLoaded });
        }
    }

    function themesModuleLoaded() {
        console.log("BingTheme Loaded");
        if (window.BingMapViewJavaScriptBridge) {
            window.BingMapViewJavaScriptBridge.onLoadScriptsFinished();
        }
    }

    //
    // Handles error loading Bing Maps AJAX SDK scripts
    //
    this.onScriptLoadError = function() {
        console.log("SDK script onError");
        scriptLoadFailed = true;
        if (window.BingMapViewJavaScriptBridge) {
            window.BingMapViewJavaScriptBridge.onLoadScriptsFailed();
        }
    }
}

